import arr
import numpy as np
import math
import scipy.signal

cfg = {
    'name': 'ECG - Heart Rate Variability',
    'group': 'Medical algorithms',
    'desc': 'Calculate Heart Rate Variability. Approximately 60-second data is required for calculating HF component and 120-second for LF. To calculate VLF, a longer signal is needed.',
    'reference': 'Heart rate variability. Standards of measurement, physiological interpretation, and clinical use. European Heart Journal (1996)17,354-381',
    'overlap': 270,  # 2 sec overlap for HR=30
    'interval': 300,  # 5 min
    'inputs': [{'name': 'ecg', 'type': 'wav'}],
    'outputs': [
        {'name': 'SDNN', 'type': 'num', 'unit': 'ms', 'min': 0, 'max': 100},
        {'name': 'RMSSD', 'type': 'num', 'unit': 'ms', 'min': 0, 'max': 10},
        {'name': 'pNN50', 'type': 'num', 'unit': '%', 'min': 0, 'max': 5},
        {'name': 'NNI', 'type': 'num', 'unit': 'ms', 'min': 500, 'max': 2500},
        {'name': 'TP', 'type': 'num', 'unit': 'ms2', 'min': 0, 'max': 1000},
        {'name': 'VLF', 'type': 'num', 'unit': 'ms2', 'min': 0, 'max': 1000},
        {'name': 'LF', 'type': 'num', 'unit': 'ms2', 'min': 0, 'max': 1000},
        {'name': 'HF', 'type': 'num', 'unit': 'ms2', 'min': 0, 'max': 1000},
        {'name': 'LFnu', 'type': 'num', 'unit': '', 'min': 0, 'max': 100},
        {'name': 'HFnu', 'type': 'num', 'unit': '', 'min': 0, 'max': 100},
        {'name': 'LF_HF', 'type': 'num', 'unit': '', 'min': 0, 'max': 10}
        ]
}


def run(inp, opt, cfg):
    data = arr.interp_undefined(inp['ecg']['vals'])
    srate = inp['ecg']['srate']

    rlist = arr.detect_qrs(data, srate)  # detect r-peaks

    # remove qrs before and after overlap
    new_rlist = []
    for ridx in rlist:
        if cfg['overlap'] <= ridx / srate:
            new_rlist.append(ridx)
    rlist = new_rlist

    ret_rpeak = [{'dt': ridx / srate, 'val': 1} for ridx in rlist]

    # average qrs
    qrs_width = int(0.1 * srate)
    qrslist = []
    for ridx in rlist:
        qrslist.append(data[ridx - qrs_width: ridx + qrs_width])
    avg_qrs = np.mean(np.array(qrslist), axis=0)

    # correlation coefficient
    celist = []
    for qrs in qrslist:
        ce = arr.corr(qrs, avg_qrs)
        celist.append(ce)

    # rr interval (ms)
    rri_list = np.diff(rlist) / srate * 1000

    nni_list = []  # nn interval (ms)
    ret_nni = []
    for i in range(len(rlist) - 1):
        if celist[i] < 0.9 or celist[i+1] < 0.9:
            continue

        # median RR interval nearest 10 beats
        med_rri = np.median(rri_list[max(0, i-5): min(len(rri_list), i+5)])
        rri = rri_list[i]
        if med_rri * 0.5 <= rri <= med_rri * 1.5:
            nni_list.append(rri)
            ret_nni.append({'dt': rlist[i+1] / srate, 'val': rri})

    # We are only interested in the region below 0.4 Hz 
    # so 7Hz will be sufficient
    nni_srate = 7
    last_nni_dt = ret_nni[-1]['dt']
    first_nni_dt = ret_nni[0]['dt']
    nni_len = int((last_nni_dt - first_nni_dt) * nni_srate) + 1
    nni_data = [None] * nni_len
    for nni in ret_nni: # ms
        nni_data[int((nni['dt'] - first_nni_dt) * nni_srate)] = nni['val']

    # make time domain nni_data function by linear interpolation
    nni_data = arr.interp_undefined(nni_data)

    # psd
    # 7hz * 300 sec = 2100
    freqs, psd = scipy.signal.welch(nni_data, nni_srate, 'hamming', 2048)
    # psd [ms2/hz]

    # integrate psd
    vlf_idx = np.logical_and(freqs >= 0.003, freqs < 0.04)
    lf_idx = np.logical_and(freqs >= 0.04, freqs < 0.15)
    hf_idx = np.logical_and(freqs >= 0.15, freqs < 0.4)
    lf = np.trapz(psd[lf_idx], freqs[lf_idx])
    hf = np.trapz(psd[hf_idx], freqs[hf_idx])
    vlf = np.trapz(psd[vlf_idx], freqs[vlf_idx])
    tp = vlf + lf + hf

    lf_nu = 0
    hf_nu = 0
    if tp - vlf:
        lf_nu = lf / (tp-vlf)
        hf_nu = hf / (tp-vlf)

    lf_hf = 0
    if hf:
        lf_hf = lf / hf

    sdnn = np.std(nni_list)

    dnni_list = abs(np.diff(nni_list))  # Difference between adjacent nn intervals
    nn50 = 0
    ret_dnni = []
    for i in range(len(dnni_list)):
        dnni = dnni_list[i]
        if dnni > 50:
            nn50 += 1
        ret_dnni.append({'dt': ret_nni[i+1]['dt'], 'val': dnni})

    pnn50 = nn50 * 100 / len(dnni_list)

    rmssdnni = 0
    if len(dnni_list) > 0:
        for dnni in dnni_list:
            rmssdnni += dnni * dnni
        rmssdnni = (rmssdnni / len(dnni_list)) ** 0.5

    dt_last = cfg['interval']
    return [
        [{'dt': dt_last, 'val': sdnn}],
        [{'dt': dt_last, 'val': rmssdnni}],
        [{'dt': dt_last, 'val': pnn50}],
        ret_nni,
        [{'dt': dt_last, 'val': tp}],
        [{'dt': dt_last, 'val': vlf}],
        [{'dt': dt_last, 'val': lf}],
        [{'dt': dt_last, 'val': hf}],
        [{'dt': dt_last, 'val': lf_nu}],
        [{'dt': dt_last, 'val': hf_nu}],
        [{'dt': dt_last, 'val': lf_hf}]
        ]
